@@ -1,5 +1,5 @@
 /*  GNU ddrescuelog - Tool for ddrescue mapfiles
-    Copyright (C) 2011-2015 Antonio Diaz Diaz.
+    Copyright (C) 2011-2019 Antonio Diaz Diaz.
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -42,15 +42,18 @@ const char * const Program_name = "GNU ddrescuelog";
 const char * const program_name = "ddrescuelog";
 const char * invocation_name = 0;
 
-enum Mode { m_none, m_and, m_change, m_compare, m_complete, m_create,
-            m_delete, m_done_st, m_invert, m_list, m_or, m_status, m_xor };
+enum Mode { m_none, m_and, m_annotate, m_change, m_compare, m_complete,
+            m_create, m_delete, m_done_st, m_invert, m_list, m_or,
+            m_shift, m_status, m_xor };
 
 
 void show_help( const int hardbs )
   {
-  std::printf( "%s - Tool for ddrescue mapfiles.\n", Program_name );
-  std::printf( "Manipulates ddrescue mapfiles, shows their contents, converts them to/from\n"
-               "other formats, compares them, and tests rescue status.\n"
+  std::printf( "GNU ddrescuelog is a tool that manipulates ddrescue mapfiles, shows mapfile\n"
+               "contents, converts mapfiles to/from other formats, compares mapfiles, tests\n"
+               "rescue status, and can delete a mapfile if the rescue is done. Ddrescuelog\n"
+               "operations can be restricted to one or several parts of the mapfile if the\n"
+               "domain setting options are used.\n"
                "\nNOTE: In versions of ddrescue prior to 1.20 the mapfile was called\n"
                "'logfile'. The format is the same; only the name has changed.\n"
                "\nUsage: %s [options] mapfile\n", invocation_name );
@@ -58,6 +61,7 @@ void show_help( const int hardbs )
                "  -h, --help                      display this help and exit\n"
                "  -V, --version                   output version information and exit\n"
                "  -a, --change-types=<ot>,<nt>    change the block types of mapfile\n"
+               "  -A, --annotate-mapfile          add comments with human-readable pos/sizes\n"
                "  -b, --block-size=<bytes>        block size in bytes [default %d]\n", hardbs );
   std::printf( "  -B, --binary-prefixes           show binary multipliers in numbers [SI]\n"
                "  -c, --create-mapfile[=<tt>]     create mapfile from list of blocks [+-]\n"
@@ -80,6 +84,9 @@ void show_help( const int hardbs )
                "  -x, --xor-mapfile=<file>        XOR the finished blocks in file with mapfile\n"
                "  -y, --and-mapfile=<file>        AND the finished blocks in file with mapfile\n"
                "  -z, --or-mapfile=<file>         OR the finished blocks in file with mapfile\n"
+               "      --shift                     shift all block positions by (opos - ipos)\n"
+               "Use '-' to read a mapfile from standard input or to write the mapfile\n"
+               "created by '--create-mapfile' to standard output.\n"
                "Numbers may be in decimal, hexadecimal or octal, and may be followed by a\n"
                "multiplier: s = sectors, k = 1000, Ki = 1024, M = 10^6, Mi = 2^20, etc...\n"
                "\nExit status: 0 for a normal exit, 1 for environmental problems (file\n"
@@ -199,7 +206,21 @@ int do_logic_ops( Domain & domain, const char * const mapname,
   mapfile.compact_sblock_vector();
   mapfile.write_mapfile( stdout );
   if( std::fclose( stdout ) != 0 )
-    { show_error( "Can't close stdout", errno ); return 1; }
+    { show_error( "Error closing stdout", errno ); return 1; }
+  return 0;
+  }
+
+
+int annotate_mapfile( Domain & domain, const char * const mapname )
+  {
+  Mapfile mapfile( mapname );
+  if( !mapfile.read_mapfile() ) return not_readable( mapname );
+  domain.crop( mapfile.extent() );
+  if( domain.empty() ) return empty_domain();
+  mapfile.split_by_domain_borders( domain );
+  mapfile.write_mapfile( stdout, false, false, &domain );
+  if( std::fclose( stdout ) != 0 )
+    { show_error( "Error closing stdout", errno ); return 1; }
   return 0;
   }
 
@@ -225,7 +246,7 @@ int change_types( Domain & domain, const char * const mapname,
   mapfile.compact_sblock_vector();
   mapfile.write_mapfile( stdout );
   if( std::fclose( stdout ) != 0 )
-    { show_error( "Can't close stdout", errno ); return 1; }
+    { show_error( "Error closing stdout", errno ); return 1; }
   return 0;
   }
 
@@ -297,7 +318,7 @@ int complete_mapfile( const char * const mapname,
   mapfile.compact_sblock_vector();
   mapfile.write_mapfile( stdout );
   if( std::fclose( stdout ) != 0 )
-    { show_error( "Can't close stdout", errno ); return 1; }
+    { show_error( "Error closing stdout", errno ); return 1; }
   return 0;
   }
 
@@ -340,7 +361,7 @@ int create_mapfile( Domain & domain, const char * const mapname,
   mapfile.truncate_vector( domain.end(), true );
   if( !mapfile.write_mapfile( to_stdout ? stdout : 0 ) ) return 1;
   if( to_stdout && std::fclose( stdout ) != 0 )
-    { show_error( "Can't close stdout", errno ); return 1; }
+    { show_error( "Error closing stdout", errno ); return 1; }
   return 0;
   }
 
@@ -385,6 +406,26 @@ int test_if_done( Domain & domain, const char * const mapname, const bool del )
   }
 
 
+int shift_blocks( const long long ipos, const long long opos,
+                  Domain & domain, const char * const mapname )
+  {
+  if( ipos != 0 && opos != 0 )
+    { show_error( "Either '-i' or '-o' must be 0" ); return 1; }
+  const long long offset = opos - ipos;
+  Mapfile mapfile( mapname );
+  if( !mapfile.read_mapfile() ) return not_readable( mapname );
+  domain.crop( mapfile.extent() );
+  if( domain.empty() ) return empty_domain();
+  mapfile.truncate_vector( domain.end(), true );
+  mapfile.shift_blocks( offset );
+  mapfile.compact_sblock_vector();
+  mapfile.write_mapfile( stdout );
+  if( std::fclose( stdout ) != 0 )
+    { show_error( "Error closing stdout", errno ); return 1; }
+  return 0;
+  }
+
+
 int to_badblocks( const long long offset, Domain & domain,
                   const char * const mapname, const int hardbs,
                   const std::string & blocktypes )
@@ -417,57 +458,15 @@ int to_badblocks( const long long offset, Domain & domain,
   }
 
 
-// Shows the fraction "num/den" as a percentage with "prec" decimals.
-// If 'prec' is negative, only the needed decimals are shown.
-//
-const char * format_percentage( long long num, long long den,
-                                const int iwidth = 3, int prec = -2 )
+int do_show_status( Domain & domain, const char * const mapname,
+                    const bool loose )
   {
-  static char buf[80];
-
-  if( den < 0 ) { num = -num; den = -den; }
-  if( llabs( num ) <= LLONG_MAX / 100 && den <= LLONG_MAX / 10 ) num *= 100;
-  else if( llabs( num ) <= LLONG_MAX / 10 ) { num *= 10; den /= 10; }
-  else den /= 100;
-  if( den == 0 )
-    {
-    if( num > 0 ) return "+INF";
-    else if( num < 0 ) return "-INF";
-    else return "NAN";
-    }
-  const bool trunc = ( prec < 0 );
-  if( prec < 0 ) prec = -prec;
-
-  unsigned i;
-  if( num < 0 && num / den == 0 )
-    i = snprintf( buf, sizeof( buf ), "%*s", iwidth, "-0" );
-  else i = snprintf( buf, sizeof( buf ), "%*lld", iwidth, num / den );
-  if( i < sizeof( buf ) - 2 )
-    {
-    long long rest = llabs( num ) % den;
-    if( prec > 0 && ( rest > 0 || !trunc ) )
-      {
-      buf[i++] = '.';
-      while( prec > 0 && ( rest > 0 || !trunc ) && i < sizeof( buf ) - 2 )
-        { rest *= 10; buf[i++] = (char)( rest / den ) + '0';
-          rest %= den; --prec; }
-      }
-    }
-  else i = sizeof( buf ) - 2;
-  buf[i++] = '%';
-  buf[i] = 0;
-  return buf;
-  }
-
-
-int do_show_status( Domain & domain, const char * const mapname )
-  {
-  long long size_non_tried = 0, size_non_trimmed = 0, size_non_scraped = 0;
-  long long size_bad_sector = 0, size_finished = 0;
-  long areas_non_tried = 0, areas_non_trimmed = 0, areas_non_scraped = 0;
-  long areas_bad_sector = 0, areas_finished = 0;
+  long long non_tried_size = 0, non_trimmed_size = 0;
+  long long non_scraped_size = 0, bad_size = 0, finished_size = 0;
+  unsigned long non_tried_areas = 0, non_trimmed_areas = 0;
+  unsigned long non_scraped_areas = 0, bad_areas = 0, finished_areas = 0;
   Mapfile mapfile( mapname );
-  if( !mapfile.read_mapfile() ) return not_readable( mapname );
+  if( !mapfile.read_mapfile( loose ? '?' : 0 ) ) return not_readable( mapname );
   mapfile.compact_sblock_vector();
   const Block extent = mapfile.extent();
   domain.crop( extent );
@@ -482,48 +481,48 @@ int do_show_status( Domain & domain, const char * const mapname )
       { if( domain < sb ) break; else continue; }
     switch( sb.status() )
       {
-      case Sblock::non_tried:   size_non_tried += sb.size();
-                                ++areas_non_tried; break;
-      case Sblock::finished:    size_finished += sb.size();
-                                ++areas_finished; break;
-      case Sblock::non_trimmed: size_non_trimmed += sb.size();
-                                ++areas_non_trimmed; break;
-      case Sblock::non_scraped: size_non_scraped += sb.size();
-                                ++areas_non_scraped; break;
-      case Sblock::bad_sector:  size_bad_sector += sb.size();
-                                ++areas_bad_sector; break;
+      case Sblock::non_tried:   non_tried_size += sb.size();
+                                ++non_tried_areas; break;
+      case Sblock::non_trimmed: non_trimmed_size += sb.size();
+                                ++non_trimmed_areas; break;
+      case Sblock::non_scraped: non_scraped_size += sb.size();
+                                ++non_scraped_areas; break;
+      case Sblock::bad_sector:  bad_size += sb.size();
+                                ++bad_areas; break;
+      case Sblock::finished:    finished_size += sb.size();
+                                ++finished_areas; break;
       }
     }
 
   const long long domain_size = domain.in_size();
   if( verbosity >= 1 ) std::printf( "\n%s", mapname );
-  std::printf( "\n   current pos: %10sB,  current status: %s\n",
+  std::printf( "\n   current pos: %9sB,  current status: %s\n",
                format_num( mapfile.current_pos() ),
                mapfile.status_name( mapfile.current_status() ) );
-  std::printf( "mapfile extent: %10sB,  in %6ld area(s)\n",
+  std::printf( "mapfile extent: %9sB,  in %6ld area(s)\n",
                format_num( extent.size() ), true_sblocks );
   if( domain.pos() > 0 || domain.end() < extent.end() || domain.blocks() > 1 )
     {
-    std::printf( "  domain begin: %10sB,  domain end: %10sB\n",
+    std::printf( "  domain begin: %9sB,  domain end: %9sB\n",
                  format_num( domain.pos() ), format_num( domain.end() ) );
-    std::printf( "   domain size: %10sB,  in %6ld area(s)\n",
+    std::printf( "   domain size: %9sB,  in %6ld area(s)\n",
                  format_num( domain_size ), domain.blocks() );
     }
-  std::printf( "\n       rescued: %10sB,  in %6ld area(s)  (%s)\n",
-               format_num( size_finished ), areas_finished,
-               format_percentage( size_finished, domain_size ) );
-  std::printf( "     non-tried: %10sB,  in %6ld area(s)  (%s)\n",
-               format_num( size_non_tried ), areas_non_tried,
-               format_percentage( size_non_tried, domain_size ) );
-  std::printf( "   non-trimmed: %10sB,  in %6ld area(s)  (%s)\n",
-               format_num( size_non_trimmed ), areas_non_trimmed,
-               format_percentage( size_non_trimmed, domain_size ) );
-  std::printf( "   non-scraped: %10sB,  in %6ld area(s)  (%s)\n",
-               format_num( size_non_scraped ), areas_non_scraped,
-               format_percentage( size_non_scraped, domain_size ) );
-  std::printf( "       errsize: %10sB,  errors:  %8ld  (%s)\n",
-               format_num( size_bad_sector ), areas_bad_sector,
-               format_percentage( size_bad_sector, domain_size ) );
+  std::printf( "\n     non-tried: %9sB,  in %6lu area(s)  (%s)\n",
+               format_num( non_tried_size ), non_tried_areas,
+               format_percentage( non_tried_size, domain_size ) );
+  std::printf( "       rescued: %9sB,  in %6lu area(s)  (%s)\n",
+               format_num( finished_size ), finished_areas,
+               format_percentage( finished_size, domain_size ) );
+  std::printf( "   non-trimmed: %9sB,  in %6lu area(s)  (%s)\n",
+               format_num( non_trimmed_size ), non_trimmed_areas,
+               format_percentage( non_trimmed_size, domain_size ) );
+  std::printf( "   non-scraped: %9sB,  in %6lu area(s)  (%s)\n",
+               format_num( non_scraped_size ), non_scraped_areas,
+               format_percentage( non_scraped_size, domain_size ) );
+  std::printf( "    bad-sector: %9sB,  in %6lu area(s)  (%s)\n",
+               format_num( bad_size ), bad_areas,
+               format_percentage( bad_size, domain_size ) );
   return 0;
   }
 
@@ -554,9 +553,11 @@ int main( const int argc, const char * const argv[] )
   for( int i = 1; i < argc; ++i )
     { command_line += ' '; command_line += argv[i]; }
 
+  enum Optcode { opt_shi = 256 };
   const Arg_parser::Option options[] =
     {
     { 'a', "change-types",        Arg_parser::yes },
+    { 'A', "annotate-mapfile",    Arg_parser::no  },
     { 'b', "block-size",          Arg_parser::yes },
     { 'b', "sector-size",         Arg_parser::yes },
     { 'B', "binary-prefixes",     Arg_parser::no  },
@@ -591,6 +592,7 @@ int main( const int argc, const char * const argv[] )
     { 'y', "and-logfile",         Arg_parser::yes },
     { 'z', "or-mapfile",          Arg_parser::yes },
     { 'z', "or-logfile",          Arg_parser::yes },
+    { opt_shi, "shift",           Arg_parser::no  },
     {  0 , 0,                     Arg_parser::no  } };
 
   const Arg_parser parser( argc, argv, options );
@@ -602,12 +604,14 @@ int main( const int argc, const char * const argv[] )
     {
     const int code = parser.code( argind );
     if( !code ) break;					// no more options
-    const char * const arg = parser.argument( argind ).c_str();
+    const std::string & arg = parser.argument( argind );
+    const char * const ptr = arg.c_str();
     switch( code )
       {
       case 'a': set_mode( program_mode, m_change );
                 parse_types( arg, types1, types2 ); break;
-      case 'b': hardbs = getnum( arg, 0, 1, INT_MAX ); break;
+      case 'A': set_mode( program_mode, m_annotate ); break;
+      case 'b': hardbs = getnum( ptr, 0, 1, INT_MAX ); break;
       case 'B': format_num( 0, 0, -1 ); break;		// set binary prefixes
       case 'c': set_mode( program_mode, m_create );
                 parse_2types( arg, type1, type2 ); break;
@@ -617,27 +621,28 @@ int main( const int argc, const char * const argv[] )
       case 'D': set_mode( program_mode, m_done_st ); break;
       case 'f': force = true; break;
       case 'h': show_help( default_hardbs ); return 0;
-      case 'i': ipos = getnum( arg, hardbs, 0 ); break;
+      case 'i': ipos = getnum( ptr, hardbs, 0 ); break;
       case 'l': set_mode( program_mode, m_list ); types1 = arg;
                 check_types( types1, "list-blocks" ); break;
       case 'L': loose = true; break;
-      case 'm': set_name( &domain_mapfile_name, arg, code ); break;
+      case 'm': set_name( &domain_mapfile_name, ptr, code ); break;
       case 'n': set_mode( program_mode, m_invert ); break;
-      case 'o': opos = getnum( arg, hardbs, 0 ); break;
+      case 'o': opos = getnum( ptr, hardbs, 0 ); break;
       case 'p':
       case 'P': set_mode( program_mode, m_compare );
-                second_mapname = arg; as_domain = ( code == 'P' ); break;
+                second_mapname = ptr; as_domain = ( code == 'P' ); break;
       case 'q': verbosity = -1; break;
-      case 's': max_size = getnum( arg, hardbs, -1 ); break;
+      case 's': max_size = getnum( ptr, hardbs, -1 ); break;
       case 't': set_mode( program_mode, m_status ); break;
       case 'v': if( verbosity < 4 ) ++verbosity; break;
       case 'V': show_version(); return 0;
       case 'x': set_mode( program_mode, m_xor );
-                second_mapname = arg; break;
+                second_mapname = ptr; break;
       case 'y': set_mode( program_mode, m_and );
-                second_mapname = arg; break;
+                second_mapname = ptr; break;
       case 'z': set_mode( program_mode, m_or );
-                second_mapname = arg; break;
+                second_mapname = ptr; break;
+      case opt_shi: set_mode( program_mode, m_shift ); break;
       default : internal_error( "uncaught option." );
       }
     } // end process options
@@ -678,6 +683,7 @@ int main( const int argc, const char * const argv[] )
       case m_or:
       case m_xor:
         return do_logic_ops( domain, mapname, second_mapname, program_mode );
+      case m_annotate: return annotate_mapfile( domain, mapname );
       case m_change: return change_types( domain, mapname, types1, types2 );
       case m_compare: return compare_mapfiles( domain, mapname, second_mapname,
                                                as_domain, loose );
@@ -689,8 +695,10 @@ int main( const int argc, const char * const argv[] )
       case m_invert: return change_types( domain, mapname, "?*/-+", "++++-" );
       case m_list:
         return to_badblocks( opos - ipos, domain, mapname, hardbs, types1 );
+      case m_shift:
+        return shift_blocks( ipos, opos, domain, mapname );
       case m_status:
-        retval = std::max( retval, do_show_status( domain, mapname ) );
+        retval = std::max( retval, do_show_status( domain, mapname, loose ) );
       }
     }
   return retval;
