@@ -1,5 +1,5 @@
 /*  GNU ddrescue - Data recovery tool
-    Copyright (C) 2004-2017 Antonio Diaz Diaz.
+    Copyright (C) 2004-2019 Antonio Diaz Diaz.
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -35,12 +35,12 @@
 
 namespace {
 
-void input_pos_error( const long long pos, const long long isize )
+void input_pos_error( const long long pos, const long long insize )
   {
   char buf[128];
   snprintf( buf, sizeof buf,
             "Can't start reading at pos %lld.\n"
-            "          Input file is only %lld bytes long.", pos, isize );
+            "          Input file is only %lld bytes long.", pos, insize );
   show_error( buf );
   }
 
@@ -84,15 +84,16 @@ bool Mapbook::emergency_save()
   }
 
 
-Mapbook::Mapbook( const long long offset, const long long isize,
+Mapbook::Mapbook( const long long offset, const long long insize,
                   Domain & dom, const Mb_options & mb_opts,
                   const char * const mapname, const int cluster,
                   const int hardbs, const bool complete_only )
   : Mapfile( mapname ), Mb_options( mb_opts ), offset_( offset ),
-    mapfile_isize_( 0 ), domain_( dom ), hardbs_( hardbs ),
+    mapfile_insize_( 0 ), domain_( dom ), hardbs_( hardbs ),
     softbs_( cluster * hardbs_ ),
     iobuf_size_( softbs_ + hardbs_ ),	// +hardbs for direct unaligned reads
-    final_errno_( 0 ), um_t1( 0 ), um_t1s( 0 ), mapfile_exists_( false )
+    final_errno_( 0 ), um_t1( 0 ), um_t1s( 0 ), um_prev_mf_sync( false ),
+    mapfile_exists_( false )
   {
   long alignment = sysconf( _SC_PAGESIZE );
   if( alignment < hardbs_ || alignment % hardbs_ ) alignment = hardbs_;
@@ -105,18 +106,18 @@ Mapbook::Mapbook( const long long offset, const long long isize,
     if( disp > 0 && disp < alignment ) iobuf_ += disp;
     }
 
-  if( isize > 0 )
+  if( insize > 0 )
     {
-    if( domain_.pos() >= isize )
-      { input_pos_error( domain_.pos(), isize ); std::exit( 1 ); }
-    domain_.crop_by_file_size( isize );
+    if( domain_.pos() >= insize )
+      { input_pos_error( domain_.pos(), insize ); std::exit( 1 ); }
+    domain_.crop_by_file_size( insize );
     }
   if( filename() )
     {
     mapfile_exists_ = read_mapfile( 0, false );
-    if( mapfile_exists_ ) mapfile_isize_ = extent().end();
+    if( mapfile_exists_ ) mapfile_insize_ = extent().end();
     }
-  if( !complete_only ) extend_sblock_vector( isize );
+  if( !complete_only ) extend_sblock_vector( insize );
   else domain_.crop( extent() );  // limit domain to blocks read from mapfile
   compact_sblock_vector();
   split_by_domain_borders( domain_ );
@@ -139,6 +140,13 @@ bool Mapbook::update_mapfile( const int odes, const bool force )
   const bool mf_sync = ( force || t2 - um_t1s >= mapfile_sync_interval );
   if( mf_sync ) um_t1s = t2;
   if( odes >= 0 ) fsync( odes );
+  if( um_prev_mf_sync )
+    {
+    std::string mapname_bak( filename() ); mapname_bak += ".bak";
+    std::remove( mapname_bak.c_str() );		// break possible hard link
+    std::rename( filename(), mapname_bak.c_str() );
+    }
+  um_prev_mf_sync = mf_sync;
 
   while( true )
     {
