@@ -1,5 +1,5 @@
 /*  GNU ddrescue - Data recovery tool
-    Copyright (C) 2004-2015 Antonio Diaz Diaz.
+    Copyright (C) 2004-2019 Antonio Diaz Diaz.
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -32,10 +32,10 @@
 
 namespace {
 
-int my_fgetc( FILE * const f )
+int my_fgetc( FILE * const f, const bool allow_comment = true )
   {
   int ch = std::fgetc( f );
-  if( ch == '#' )			// comment
+  if( ch == '#' && allow_comment )			// comment
     { do ch = std::fgetc( f ); while( ch != '\n' && ch != EOF ); }
   return ch;
   }
@@ -60,7 +60,7 @@ const char * my_fgets( FILE * const f, int & linenum )
       if( ch == EOF ) { if( len > 0 ) ch = '\n'; else break; }
       if( len < maxlen ) buf[len++] = ch;
       if( ch == '\n' ) { ++linenum; break; }
-      ch = my_fgetc( f );
+      ch = my_fgetc( f, std::isspace( ch ) );
       }
     }
   if( len > 0 ) { buf[len] = 0; return buf; }
@@ -80,9 +80,12 @@ void show_mapfile_error( const char * const mapname, const int linenum )
 
 void Mapfile::compact_sblock_vector()
   {
+  unsigned long l;
+  for( l = 1; l < sblock_vector.size(); ++l )
+    if( sblock_vector[l-1].status() == sblock_vector[l].status() ) break;
+  if( l >= sblock_vector.size() ) return;	// already compacted
   std::vector< Sblock > new_vector;
-  unsigned long l = 0;
-  while( l < sblock_vector.size() )
+  for( l = 0; l < sblock_vector.size(); )
     {
     Sblock run = sblock_vector[l];
     unsigned long r = l + 1;
@@ -96,11 +99,11 @@ void Mapfile::compact_sblock_vector()
   }
 
 
-void Mapfile::extend_sblock_vector( const long long isize )
+void Mapfile::extend_sblock_vector( const long long insize )
   {
   if( sblock_vector.empty() )
     {
-    const Sblock sb( 0, ( isize > 0 ) ? isize : -1, Sblock::non_tried );
+    const Sblock sb( 0, ( insize > 0 ) ? insize : -1, Sblock::non_tried );
     sblock_vector.push_back( sb );
     return;
     }
@@ -109,28 +112,28 @@ void Mapfile::extend_sblock_vector( const long long isize )
     sblock_vector.insert( sblock_vector.begin(), Sblock( 0, front.pos(), Sblock::non_tried ) );
   Sblock & back = sblock_vector.back();
   const long long end = back.end();
-  if( isize > 0 )
+  if( insize > 0 )
     {
-    if( back.pos() >= isize )
+    if( back.pos() >= insize )
       {
-      if( back.pos() == isize && back.status() != Sblock::finished )
+      if( back.pos() == insize && back.status() != Sblock::finished )
         { sblock_vector.pop_back(); return; }
       show_error( "Last block in mapfile begins past end of input file.\n"
                   "          Use '-C' if you are reading from a partial copy.",
                   0, true );
       std::exit( 1 );
       }
-    if( end > isize )
+    if( end > insize )
       {
       if( back.status() != Sblock::finished )
-        { back.size( isize - back.pos() ); return; }
+        { back.size( insize - back.pos() ); return; }
       show_error( "Rescued data in mapfile goes past end of input file.\n"
                   "          Use '-C' if you are reading from a partial copy.",
                   0, true );
       std::exit( 1 );
       }
-    else if( end < isize )
-      sblock_vector.push_back( Sblock( end, isize - end, Sblock::non_tried ) );
+    else if( end < insize )
+      sblock_vector.push_back( Sblock( end, insize - end, Sblock::non_tried ) );
     }
   else if( end >= 0 )
     {
@@ -140,6 +143,38 @@ void Mapfile::extend_sblock_vector( const long long isize )
   }
 
 
+void Mapfile::shift_blocks( const long long offset )
+  {
+  if( sblock_vector.empty() ) return;
+  if( offset > 0 )
+    {
+    if( sblock_vector.front().status() == Sblock::non_tried )
+      sblock_vector.front().enlarge( offset );
+    else
+      insert_sblock( 0, Sblock( 0, offset, Sblock::non_tried ) );
+    for( unsigned long i = 1; i < sblock_vector.size(); ++i )
+      {
+      sblock_vector[i].shift( offset );
+      if( sblock_vector[i].size() <= 0 )
+        { sblock_vector.erase( sblock_vector.begin() + i, sblock_vector.end() );
+          break; }
+      }
+    }
+  else if( offset < 0 )
+    {
+    for( unsigned long i = 0; i < sblock_vector.size(); ++i )
+      if( sblock_vector[i].end() + offset > 0 )
+        {
+        if( i > 0 )
+          sblock_vector.erase( sblock_vector.begin(), sblock_vector.begin() + i );
+        break;
+        }
+    for( unsigned long i = 0; i < sblock_vector.size(); ++i )
+      sblock_vector[i].shift( offset );
+    }
+  }
+
+
 // Returns false only if truncation would remove finished blocks and
 // force is false.
 //
@@ -177,8 +212,7 @@ bool Mapfile::read_mapfile( const int default_sblock_status, const bool ro )
   FILE * f = 0;
   errno = 0;
   read_only_ = ro;
-  if( ro && std::strcmp( filename_, "-" ) == 0 && !isatty( fileno( stdin ) ) )
-    f = stdin;
+  if( ro && std::strcmp( filename_, "-" ) == 0 ) f = stdin;
   else if( ro || ( !(f = std::fopen( filename_, "r+" )) && errno != ENOENT ) )
     { f = std::fopen( filename_, "r" ); read_only_ = true; }
   if( !f ) return false;
@@ -190,8 +224,11 @@ bool Mapfile::read_mapfile( const int default_sblock_status, const bool ro )
   if( line )						// status line
     {
     char ch;
-    int n = std::sscanf( line, "%lli %c\n", &current_pos_, &ch );
-    if( n == 2 && current_pos_ >= 0 && isstatus( ch ) )
+    current_pass_ = 1;					// default value
+    const int n = std::sscanf( line, "%lli %c %d\n",
+                               &current_pos_, &ch, &current_pass_ );
+    if( ( n == 3 || n == 2 ) && current_pos_ >= 0 && isstatus( ch ) &&
+        current_pass_ >= 1 )
       current_status_ = Status( ch );
     else
       { show_mapfile_error( filename_, linenum ); std::exit( 2 ); }
@@ -201,12 +238,11 @@ bool Mapfile::read_mapfile( const int default_sblock_status, const bool ro )
       line = my_fgets( f, linenum );
       if( !line ) break;
       long long pos, size;
-      n = std::sscanf( line, "%lli %lli %c\n", &pos, &size, &ch );
+      const int n = std::sscanf( line, "%lli %lli %c\n", &pos, &size, &ch );
       if( n == 3 && pos >= 0 && Sblock::isstatus( ch ) &&
           ( size > 0 || ( size == 0 && pos == 0 ) ) )
         {
-        const Sblock::Status st = Sblock::Status( ch );
-        const Sblock sb( pos, size, st );
+        const Sblock sb( pos, size, Sblock::Status( ch ) );
         const long long end = sblock_vector.size() ?
                               sblock_vector.back().end() : 0;
         if( sb.pos() != end )
@@ -230,7 +266,9 @@ bool Mapfile::read_mapfile( const int default_sblock_status, const bool ro )
   }
 
 
-int Mapfile::write_mapfile( FILE * f, const bool timestamp ) const
+int Mapfile::write_mapfile( FILE * f, const bool timestamp,
+                            const bool mf_sync,
+                            const Domain * const annotate_domainp ) const
   {
   const bool f_given = ( f != 0 );
 
@@ -239,15 +277,24 @@ int Mapfile::write_mapfile( FILE * f, const bool timestamp ) const
   write_file_header( f, "Mapfile" );
   if( timestamp ) write_timestamp( f );
   if( current_msg.size() ) std::fprintf( f, "# %s\n", current_msg.c_str() );
-  std::fprintf( f, "# current_pos  current_status\n"
-                   "0x%08llX     %c\n"
+  char buf[80] = { 0 };		// comment
+  if( annotate_domainp )
+    snprintf( buf, sizeof buf, "\t#  %sB", format_num( current_pos_ ) );
+  std::fprintf( f, "# current_pos  current_status  current_pass\n"
+                   "0x%08llX     %c               %d%s\n"
                    "#      pos        size  status\n",
-                current_pos_, current_status_ );
+                current_pos_, current_status_, current_pass_, buf );
   for( unsigned long i = 0; i < sblock_vector.size(); ++i )
     {
     const Sblock & sb = sblock_vector[i];
-    std::fprintf( f, "0x%08llX  0x%08llX  %c\n", sb.pos(), sb.size(), sb.status() );
+    if( annotate_domainp && annotate_domainp->includes( sb ) )
+      snprintf( buf, sizeof buf, "\t#  %9sB  %9s%c", format_num( sb.pos() ),
+                format_num( sb.size() ), ( sb.size() > 999999 ) ? 'B' : ' ' );
+    else buf[0] = 0;
+    std::fprintf( f, "0x%08llX  0x%08llX  %c%s\n",
+                  sb.pos(), sb.size(), sb.status(), buf );
     }
+  if( mf_sync ) fsync( fileno( f ) );
   return ( f_given || std::fclose( f ) == 0 );
   }
 
@@ -332,56 +379,81 @@ long Mapfile::find_index( const long long pos ) const
   }
 
 
-// Find chunk from b.pos of size <= b.size and status st.
-// If not found, put b.size to 0.
+// Find chunk from b.pos forwards of size <= b.size and status st.
+// if unfinished is true, find also any chunk not marked as finished.
+// If not found, or if after_finished is true and none of the blocks
+// found follows a finished block, put b.size to 0.
+// If at least one block of status st is found, return true.
 //
-void Mapfile::find_chunk( Block & b, const Sblock::Status st,
-                          const Domain & domain, const int alignment ) const
+bool Mapfile::find_chunk( Block & b, const Sblock::Status st,
+                          const Domain & domain, const int alignment,
+                          const bool after_finished,
+                          const bool unfinished ) const
   {
-  if( b.size() <= 0 ) return;
+  if( b.size() <= 0 ) return false;
   if( b.pos() < sblock_vector.front().pos() )
     b.pos( sblock_vector.front().pos() );
-  if( find_index( b.pos() ) < 0 ) { b.size( 0 ); return; }
+  if( find_index( b.pos() ) < 0 ) { b.size( 0 ); return false; }
   long i;
+  bool block_found = false;
   for( i = index_; i < sblocks(); ++i )
-    if( sblock_vector[i].status() == st && domain.includes( sblock_vector[i] ) )
-      { index_ = i; break; }
-  if( i >= sblocks() ) { b.size( 0 ); return; }
+    {
+    const Sblock::Status ist = sblock_vector[i].status();
+    if( ( ist == st || ( unfinished && ist != Sblock::finished ) ) &&
+        domain.includes( sblock_vector[i] ) )
+      {
+      block_found = true;
+      if( !after_finished || i <= 0 ||
+          sblock_vector[i-1].status() == Sblock::finished )
+        { index_ = i; break; }
+      }
+    }
+  if( i >= sblocks() ) { b.size( 0 ); return block_found; }
   if( b.pos() < sblock_vector[index_].pos() )
     b.pos( sblock_vector[index_].pos() );
   if( !sblock_vector[index_].includes( b ) )
     b.crop( sblock_vector[index_] );
   if( b.end() != sblock_vector[index_].end() )
     b.align_end( alignment );
+  return block_found;
   }
 
 
 // Find chunk from b.end backwards of size <= b.size and status st.
-// If not found, put b.size to 0.
+// If not found, or if before_finished is true and none of the blocks
+// found precedes a finished block, put b.size to 0.
+// If at least one block of status st is found, return true.
 //
-void Mapfile::rfind_chunk( Block & b, const Sblock::Status st,
-                           const Domain & domain, const int alignment ) const
+bool Mapfile::rfind_chunk( Block & b, const Sblock::Status st,
+                           const Domain & domain, const int alignment,
+                           const bool before_finished ) const
   {
-  if( b.size() <= 0 ) return;
-  if( sblock_vector.back().end() < b.end() )
+  if( b.size() <= 0 ) return false;
+  if( b.end() > sblock_vector.back().end() )
     b.end( sblock_vector.back().end() );
-  if( find_index( b.end() - 1 ) < 0 ) { b.size( 0 ); return; }
+  if( find_index( b.end() - 1 ) < 0 ) { b.size( 0 ); return false; }
   long i;
+  bool block_found = false;
   for( i = index_; i >= 0; --i )
     if( sblock_vector[i].status() == st && domain.includes( sblock_vector[i] ) )
-      { index_ = i; break; }
-  if( i < 0 ) { b.size( 0 ); return; }
+      {
+      block_found = true;
+      if( !before_finished || i + 1 >= sblocks() ||
+          sblock_vector[i+1].status() == Sblock::finished )
+        { index_ = i; break; }
+      }
+  if( i < 0 ) { b.size( 0 ); return block_found; }
   if( b.end() > sblock_vector[index_].end() )
     b.end( sblock_vector[index_].end() );
   if( !sblock_vector[index_].includes( b ) )
     b.crop( sblock_vector[index_] );
   if( b.pos() != sblock_vector[index_].pos() )
     b.align_pos( alignment );
+  return block_found;
   }
 
 
-// Returns an adjust value (-1, 0, +1) to keep "errors" updated without
-// having to call count_errors every time.
+// Returns an adjust value (-1, 0, +1) to keep 'bad_areas' updated.
 //   - - -   -->   - + -   return +1
 //   - - +   -->   - + +   return  0
 //   - + -   -->   - - -   return -1
@@ -392,7 +464,8 @@ void Mapfile::rfind_chunk( Block & b, const Sblock::Status st,
 //   + + +   -->   + - +   return +1
 //
 int Mapfile::change_chunk_status( const Block & b, const Sblock::Status st,
-                                  const Domain & domain )
+                                  const Domain & domain,
+                                  Sblock::Status * const old_stp )
   {
   if( b.size() <= 0 ) return 0;
   if( !domain.includes( b ) || find_index( b.pos() ) < 0 ||
@@ -401,6 +474,7 @@ int Mapfile::change_chunk_status( const Block & b, const Sblock::Status st,
   if( !sblock_vector[index_].includes( b ) )
     internal_error( "can't change status of chunk spread over more than 1 block." );
   const Sblock::Status old_st = sblock_vector[index_].status();
+  if( old_stp ) *old_stp = old_st;
   if( st == old_st ) return 0;
   const bool old_st_good = Sblock::is_good_status( old_st );
   const bool new_st_good = Sblock::is_good_status( st );
@@ -417,7 +491,7 @@ int Mapfile::change_chunk_status( const Block & b, const Sblock::Status st,
         index_ + 1 < sblocks() && sblock_vector[index_+1].status() == st &&
         domain.includes( sblock_vector[index_+1] ) )
       {
-      sblock_vector[index_].shift( sblock_vector[index_+1], b.pos() );
+      sblock_vector[index_].shift_boundary( sblock_vector[index_+1], b.pos() );
       return 0;
       }
     insert_sblock( index_, sblock_vector[index_].split( b.pos() ) );
@@ -428,7 +502,7 @@ int Mapfile::change_chunk_status( const Block & b, const Sblock::Status st,
     {
     if( index_ > 0 && sblock_vector[index_-1].status() == st &&
         domain.includes( sblock_vector[index_-1] ) )
-      sblock_vector[index_-1].shift( sblock_vector[index_], b.end() );
+      sblock_vector[index_-1].shift_boundary( sblock_vector[index_], b.end() );
     else
       insert_sblock( index_,
                      Sblock( sblock_vector[index_].split( b.end() ), st ) );
