@@ -1,5 +1,5 @@
 /*  GNU ddrescue - Data recovery tool
-    Copyright (C) 2004-2016 Antonio Diaz Diaz.
+    Copyright (C) 2004-2019 Antonio Diaz Diaz.
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -30,6 +30,7 @@
 #include <unistd.h>
 #include <sys/stat.h>
 
+#include "rational.h"
 #include "block.h"
 #include "loggers.h"
 #include "mapbook.h"
@@ -40,12 +41,12 @@ namespace {
 
 // Round "size" to the next multiple of sector size (hardbs).
 //
-int round_up( int size, const int hardbs )
+long long round_up( long long size, const int hardbs )
   {
   if( size % hardbs )
     {
     size -= size % hardbs;
-    if( INT_MAX - size >= hardbs ) size += hardbs;
+    if( LLONG_MAX - size >= hardbs ) size += hardbs;
     }
   return size;
   }
@@ -56,14 +57,14 @@ int round_up( int size, const int hardbs )
 void Rescuebook::change_chunk_status( const Block & b, const Sblock::Status st )
   {
   Sblock::Status old_st = st;
-  errors += Mapfile::change_chunk_status( b, st, domain(), &old_st );
+  bad_areas += Mapfile::change_chunk_status( b, st, domain(), &old_st );
   if( st == old_st ) return;
   switch( old_st )
     {
     case Sblock::non_tried:     non_tried_size -= b.size(); break;
     case Sblock::non_trimmed: non_trimmed_size -= b.size(); break;
     case Sblock::non_scraped: non_scraped_size -= b.size(); break;
-    case Sblock::bad_sector:   bad_sector_size -= b.size(); break;
+    case Sblock::bad_sector:          bad_size -= b.size(); break;
     case Sblock::finished:       finished_size -= b.size(); break;
     }
   switch( st )
@@ -71,12 +72,19 @@ void Rescuebook::change_chunk_status( const Block & b, const Sblock::Status st )
     case Sblock::non_tried:     non_tried_size += b.size(); break;
     case Sblock::non_trimmed: non_trimmed_size += b.size(); break;
     case Sblock::non_scraped: non_scraped_size += b.size(); break;
-    case Sblock::bad_sector:   bad_sector_size += b.size(); break;
+    case Sblock::bad_sector:          bad_size += b.size(); break;
     case Sblock::finished:       finished_size += b.size(); break;
     }
   }
 
 
+void Rescuebook::do_pause_on_error()
+  {
+  if( simulated_poe ) tp += pause_on_error;
+  else if( pause_on_error >= 1 ) sleep( pause_on_error.trunc() );
+  }
+
+
 bool Rescuebook::extend_outfile_size()
   {
   if( min_outfile_size > 0 || sparse_size > 0 )
@@ -87,7 +95,7 @@ bool Rescuebook::extend_outfile_size()
     if( min_size > size )
       {
       const uint8_t zero = 0;
-      if( writeblock( odes_, &zero, 1, min_size - 1 ) != 1 ) return false;
+      if( writeblockp( odes_, &zero, 1, min_size - 1 ) != 1 ) return false;
       fsync( odes_ );
       }
     }
@@ -111,13 +119,13 @@ int Rescuebook::copy_block( const Block & b, int & copied_size, int & error_size
       const int size = pre + b.size() + post;
       if( size > iobuf_size() )
         internal_error( "(size > iobuf_size) copying a Block." );
-      copied_size = readblock( ides_, iobuf(), size, b.pos() - pre );
+      copied_size = readblockp( ides_, iobuf(), size, b.pos() - pre );
       copied_size -= std::min( pre, copied_size );
       if( copied_size > b.size() ) copied_size = b.size();
       if( pre > 0 && copied_size > 0 )
         std::memmove( iobuf(), iobuf() + pre, copied_size );
       }
-    else copied_size = readblock( ides_, iobuf(), b.size(), b.pos() );
+    else copied_size = readblockp( ides_, iobuf(), b.size(), b.pos() );
     error_size = errno ? b.size() - copied_size : 0;
     if( errno == EINVAL )
       { final_msg( "Unaligned read error. Is sector size correct?" ); return 1; }
@@ -133,7 +141,7 @@ int Rescuebook::copy_block( const Block & b, int & copied_size, int & error_size
       const long long end = pos + copied_size;
       if( end > sparse_size ) sparse_size = end;
       }
-    else if( writeblock( odes_, iobuf(), copied_size, pos ) != copied_size ||
+    else if( writeblockp( odes_, iobuf(), copied_size, pos ) != copied_size ||
              ( synchronous_ && fsync( odes_ ) != 0 && errno != EINVAL ) )
       { final_msg( "Write error", errno ); return 1; }
     }
@@ -149,7 +157,7 @@ int Rescuebook::copy_block( const Block & b, int & copied_size, int & error_size
       {
       if( voe_ipos >= 0 )
         {
-        const int size = readblock( ides_, iobuf_aux(), hardbs(), voe_ipos );
+        const int size = readblockp( ides_, iobuf_aux(), hardbs(), voe_ipos );
         if( size != hardbs() )
           { final_msg( "Input file no longer returns data", errno ); e_code |= 8; }
         else if( std::memcmp( voe_buf, iobuf_aux(), hardbs() ) != 0 )
@@ -168,8 +176,8 @@ void Rescuebook::initialize_sizes()
   {
   bool good = true;
   non_tried_size = non_trimmed_size = non_scraped_size = 0;
-  bad_sector_size = finished_size = 0;
-  errors = 0;
+  bad_size = finished_size = 0;
+  bad_areas = 0;
 
   for( long i = 0; i < sblocks(); ++i )
     {
@@ -181,33 +189,35 @@ void Rescuebook::initialize_sizes()
       case Sblock::non_tried:     non_tried_size += sb.size(); good = true; break;
       case Sblock::non_trimmed: non_trimmed_size += sb.size(); good = true; break;
       case Sblock::non_scraped: non_scraped_size += sb.size(); good = true; break;
-      case Sblock::bad_sector:   bad_sector_size += sb.size();
-        if( good ) { good = false; ++errors; } break;
+      case Sblock::bad_sector:          bad_size += sb.size();
+        if( good ) { good = false; ++bad_areas; } break;
       case Sblock::finished:       finished_size += sb.size(); good = true; break;
       }
     }
   }
 
 
-// Return values: 1 I/O error, 0 OK, -1 interrupted.
+// Return values: 1 error, 0 OK, -1 interrupted.
 //
 int Rescuebook::copy_and_update( const Block & b, int & copied_size,
                                  int & error_size, const char * const msg,
-                                 const Status curr_st, const bool forward,
-                                 const Sblock::Status st )
+                                 const Status curr_st, const int curr_pass,
+                                 const bool forward, const Sblock::Status st )
   {
   if( first_post )
     {
     if( first_read ) first_read = false;
-    else if( pause > 0 )
+    else if( pause_on_pass > 0 )
       {
       show_status( -1, "Paused", true );
-      sleep( pause );
+      sleep( pause_on_pass );
       const long t2 = std::time( 0 );
       if( t1 < t2 ) t1 = t2;			// clock may have jumped back
-      ts = std::min( ts + pause, t2 );		// avoid spurious timeout
+      ts = std::min( ts + pause_on_pass, t2 );	// avoid spurious timeout
       }
     current_status( curr_st, msg );
+    current_pass( curr_pass );
+    event_logger.print_msg( t1 - t0, percent_rescued(), msg );
     read_logger.print_msg( t1 - t0, msg );
     }
   current_pos( forward ? b.pos() : b.end() );
@@ -229,7 +239,9 @@ int Rescuebook::copy_and_update( const Block & b, int & copied_size,
       change_chunk_status( Block( b.pos(), copied_size ), Sblock::finished );
     if( error_size > 0 )
       {
-      error_rate += error_size;
+      error_sum += error_size;
+      ++read_errors;
+      if( read_errors > max_read_errors ) { e_code |= 16; retval = 1; }
       const Sblock::Status st2 =
         ( error_size > hardbs() ) ? st : Sblock::bad_sector;
       change_chunk_status( Block( b.pos() + copied_size, error_size ), st2 );
@@ -249,20 +261,25 @@ int Rescuebook::copy_non_tried()
   {
   char msgbuf[80] = "Copying non-tried blocks... Pass ";
   const int msglen = std::strlen( msgbuf );
+  const bool cpass_given = ( cpass_bitset != 31 );
+  bool resume = ( !cpass_given && current_status() == copying );
+  if( !resume || current_pass() > 5 ) current_pass( 1 );	// reset pass
+  const int first_pass = current_pass();
   bool forward = !reverse;
 
-  for( int pass = 1; pass <= 3; ++pass )
+  for( int pass = 1; pass <= 5; ++pass )
     {
-    if( cpass_bitset & ( 1 << ( pass - 1 ) ) )
+    if( pass >= first_pass && cpass_bitset & ( 1 << ( pass - 1 ) ) )
       {
+      if( pass != first_pass ) resume = false;
       first_post = true;
       snprintf( msgbuf + msglen, ( sizeof msgbuf ) - msglen, "%d %s",
                 pass, forward ? "(forwards)" : "(backwards)" );
-      int retval = forward ? fcopy_non_tried( msgbuf, pass ) :
-                             rcopy_non_tried( msgbuf, pass );
+      int retval = forward ? fcopy_non_tried( msgbuf, pass, resume ) :
+                             rcopy_non_tried( msgbuf, pass, resume );
       if( retval != -3 ) return retval;
-      reduce_min_read_rate();
       }
+    if( pass >= 2 && min_read_rate >= 0 ) min_read_rate = -1;	// reset rate
     if( !unidirectional ) forward = !forward;
     }
   return 0;
@@ -273,14 +290,16 @@ int Rescuebook::copy_non_tried()
 // Read forwards the non-tried part of the domain, skipping over the
 // damaged areas.
 //
-int Rescuebook::fcopy_non_tried( const char * const msg, const int pass )
+int Rescuebook::fcopy_non_tried( const char * const msg, const int pass,
+                                 const bool resume )
   {
   long long pos = 0;
-  int skip_size = skipbs;		// size to skip on error if skipbs > 0
+  long long eskip_size = skipbs;	// size to skip on error if skipbs > 0
+  long long sskip_size = 0;		// size to skip on slow if skipbs > 0
+  const bool after_finished = (pass == 3 || pass == 4 );
   bool block_found = false;
 
-  if( pass == 1 && current_status() == copying &&
-      domain().includes( current_pos() ) )
+  if( resume && domain().includes( current_pos() ) )
     {
     Block b( current_pos(), 1 );
     find_chunk( b, Sblock::non_tried, domain(), hardbs() );
@@ -290,30 +309,48 @@ int Rescuebook::fcopy_non_tried( const char * const msg, const int pass )
   while( pos >= 0 )
     {
     Block b( pos, softbs() );
-    find_chunk( b, Sblock::non_tried, domain(), softbs() );
+    if( find_chunk( b, Sblock::non_tried, domain(), softbs(), after_finished ) )
+      block_found = true;
     if( b.size() <= 0 ) break;
-    if( pos != b.pos() ) skip_size = skipbs;	// reset size on block change
+    if( pos != b.pos() )		// reset size on block change
+      { eskip_size = skipbs; current_slow = false; }
     pos = b.end();
-    block_found = true;
     int copied_size = 0, error_size = 0;
     const int retval = copy_and_update( b, copied_size, error_size, msg,
-                                        copying, true, Sblock::non_trimmed );
+                                        copying, pass, true, Sblock::non_trimmed );
     if( retval ) return retval;
-    update_rates();
-    if( error_size > 0 && exit_on_error ) { e_code |= 2; return 1; }
-    if( ( error_size > 0 || slow_read() ) && pos >= 0 )
+    const bool slow = update_rates();
+    if( slow )
+      { ++slow_reads;
+        if( slow_reads > max_slow_reads ) { e_code |= 32; return 1; } }
+    if( ( error_size > 0 || ( slow && pass <= 2 ) ) && pos >= 0 )
       {
       if( reopen_on_error && !reopen_infile() ) return 1;
-      if( skipbs > 0 && pass <= 2 )		// don't skip if skipbs == 0
+      if( pause_on_error > 0 ) do_pause_on_error();
+      if( skipbs > 0 && pass <= 4 )		// don't skip if skipbs == 0
         {
-        b.assign( pos, skip_size );
+        b.pos( pos );
+        if( pass >= 2 )	// skip rest of block at first error or slow read
+          b.size( -1 );
+        else if( error_size > 0 )
+          {
+          b.size( eskip_size );
+          if( eskip_size <= max_skipbs / 2 ) eskip_size *= 2;
+          else eskip_size = max_skipbs;
+          }
+        else				// slow read on pass 1
+          {
+          if( !prev_slow )
+            sskip_size = std::max( skipbs, std::min( c_rate, max_skipbs ) );
+          else if( sskip_size <= max_skipbs / 2 ) sskip_size *= 2;
+          else sskip_size = max_skipbs;
+          b.size( sskip_size );
+          }
         find_chunk( b, Sblock::non_tried, domain(), hardbs() );
         if( pos == b.pos() && b.size() > 0 ) pos = b.end();	// skip
-        if( skip_size <= max_skipbs / 2 ) skip_size *= 2;
-        else skip_size = max_skipbs;
         }
       }
-    else if( copied_size > 0 ) skip_size = skipbs;		// reset
+    else if( error_size == 0 && copied_size > 0 ) eskip_size = skipbs;	// reset
     if( !update_mapfile( odes_ ) ) return -2;
     }
   if( !block_found ) return 0;
@@ -325,14 +362,16 @@ int Rescuebook::fcopy_non_tried( const char * const msg, const int pass )
 // Read backwards the non-tried part of the domain, skipping over the
 // damaged areas.
 //
-int Rescuebook::rcopy_non_tried( const char * const msg, const int pass )
+int Rescuebook::rcopy_non_tried( const char * const msg, const int pass,
+                                 const bool resume )
   {
   long long end = LLONG_MAX;
-  int skip_size = skipbs;		// size to skip on error if skipbs > 0
+  long long eskip_size = skipbs;	// size to skip on error if skipbs > 0
+  long long sskip_size = 0;		// size to skip on slow if skipbs > 0
+  const bool before_finished = (pass == 3 || pass == 4 );
   bool block_found = false;
 
-  if( pass == 1 && current_status() == copying &&
-      domain().includes( current_pos() - 1 ) )
+  if( resume && domain().includes( current_pos() - 1 ) )
     {
     Block b( current_pos() - 1, 1 );
     rfind_chunk( b, Sblock::non_tried, domain(), hardbs() );
@@ -342,30 +381,47 @@ int Rescuebook::rcopy_non_tried( const char * const msg, const int pass )
   while( end > 0 )
     {
     Block b( end - softbs(), softbs() );
-    rfind_chunk( b, Sblock::non_tried, domain(), softbs() );
+    if( rfind_chunk( b, Sblock::non_tried, domain(), softbs(), before_finished ) )
+      block_found = true;
     if( b.size() <= 0 ) break;
-    if( end != b.end() ) skip_size = skipbs;	// reset size on block change
+    if( end != b.end() )		// reset size on block change
+      { eskip_size = skipbs; current_slow = false; }
     end = b.pos();
-    block_found = true;
     int copied_size = 0, error_size = 0;
     const int retval = copy_and_update( b, copied_size, error_size, msg,
-                                        copying, false, Sblock::non_trimmed );
+                                        copying, pass, false, Sblock::non_trimmed );
     if( retval ) return retval;
-    update_rates();
-    if( error_size > 0 && exit_on_error ) { e_code |= 2; return 1; }
-    if( ( error_size > 0 || slow_read() ) && end > 0 )
+    const bool slow = update_rates();
+    if( slow )
+      { ++slow_reads;
+        if( slow_reads > max_slow_reads ) { e_code |= 32; return 1; } }
+    if( ( error_size > 0 || ( slow && pass <= 2 ) ) && end > 0 )
       {
       if( reopen_on_error && !reopen_infile() ) return 1;
-      if( skipbs > 0 && pass <= 2 )		// don't skip if skipbs == 0
+      if( pause_on_error > 0 ) do_pause_on_error();
+      if( skipbs > 0 && pass <= 4 )		// don't skip if skipbs == 0
         {
-        b.assign( end - skip_size, skip_size );
+        if( pass >= 2 )	// skip rest of block at first error or slow read
+          b.assign( 0, end );
+        else if( error_size > 0 )
+          {
+          b.assign( end - eskip_size, eskip_size );
+          if( eskip_size <= max_skipbs / 2 ) eskip_size *= 2;
+          else eskip_size = max_skipbs;
+          }
+        else				// slow read on pass 1
+          {
+          if( !prev_slow )
+            sskip_size = std::max( skipbs, std::min( c_rate, max_skipbs ) );
+          else if( sskip_size <= max_skipbs / 2 ) sskip_size *= 2;
+          else sskip_size = max_skipbs;
+          b.assign( end - sskip_size, sskip_size );
+          }
         rfind_chunk( b, Sblock::non_tried, domain(), hardbs() );
         if( end == b.end() && b.size() > 0 ) end = b.pos();	// skip
-        if( skip_size <= max_skipbs / 2 ) skip_size *= 2;
-        else skip_size = max_skipbs;
         }
       }
-    else if( copied_size > 0 ) skip_size = skipbs;		// reset
+    else if( error_size == 0 && copied_size > 0 ) eskip_size = skipbs;	// reset
     if( !update_mapfile( odes_ ) ) return -2;
     }
   if( !block_found ) return 0;
@@ -374,7 +430,8 @@ int Rescuebook::rcopy_non_tried( const char * const msg, const int pass )
 
 
 // Return values: 1 I/O error, 0 OK, -1 interrupted, -2 mapfile error.
-// Trim both edges of each damaged area sequentially.
+// Trim both edges of each damaged area sequentially. If any edge is
+// adjacent to a bad sector, leave it for the scraping phase.
 //
 int Rescuebook::trim_errors()
   {
@@ -384,31 +441,38 @@ int Rescuebook::trim_errors()
 
   for( long i = 0; i < sblocks(); )
     {
-    const Sblock sb = sblock( reverse ? sblocks() - i - 1 : i );
+    const long idx = reverse ? sblocks() - 1 - i : i;
+    const Sblock sb = sblock( idx );
     if( !domain().includes( sb ) )
       { if( ( !reverse && domain() < sb ) || ( reverse && domain() > sb ) )
           break;
         ++i; continue; }
     if( sb.status() != Sblock::non_trimmed ) { ++i; continue; }
+    const bool lbad = ( idx > 0 &&
+                        sblock( idx - 1 ).status() == Sblock::bad_sector );
+    const bool rbad = ( idx + 1 < sblocks() &&
+                        sblock( idx + 1 ).status() == Sblock::bad_sector );
+    if( lbad && rbad )		// leave block for the scraping phase
+      { change_sblock_status( idx, Sblock::non_scraped ); ++i; continue; }
+    bool error_found = lbad;
     long long pos = sb.pos();
     long long end = sb.end();
-    bool error_found = false;
-    while( pos < end && !error_found )
+    while( pos < end && !error_found )		// trim leading edge
       {
       Block b( pos, std::min( (long long)hardbs(), end - pos ) );
       if( b.end() != end ) b.align_end( hardbs() );
       pos = b.end();
       int copied_size = 0, error_size = 0;
       const int retval = copy_and_update( b, copied_size, error_size, msg,
-                                          trimming, true );
+                                          trimming, 1, true );
       if( retval ) return retval;
       update_rates();
       if( error_size > 0 )
-        { error_found = true; if( exit_on_error ) { e_code |= 2; return 1; } }
+        { error_found = true; if( pause_on_error > 0 ) do_pause_on_error(); }
       if( !update_mapfile( odes_ ) ) return -2;
       }
-    error_found = false;
-    while( end > pos && !error_found )
+    error_found = rbad;
+    while( pos < end && !error_found )		// trim trailing edge
       {
       const int size = std::min( (long long)hardbs(), end - pos );
       Block b( end - size, size );
@@ -416,20 +480,20 @@ int Rescuebook::trim_errors()
       end = b.pos();
       int copied_size = 0, error_size = 0;
       const int retval = copy_and_update( b, copied_size, error_size, msg,
-                                          trimming, false );
+                                          trimming, 1, false );
       if( retval ) return retval;
-      if( error_size > 0 && end > pos )
-        {
-        const long index = find_index( end - 1 );
-        if( index >= 0 && domain().includes( sblock( index ) ) &&
-            sblock( index ).status() == Sblock::non_trimmed )
-          change_chunk_status( sblock( index ), Sblock::non_scraped );
-        }
       update_rates();
       if( error_size > 0 )
-        { error_found = true; if( exit_on_error ) { e_code |= 2; return 1; } }
+        { error_found = true; if( pause_on_error > 0 ) do_pause_on_error(); }
       if( !update_mapfile( odes_ ) ) return -2;
       }
+    if( pos < end )
+      {
+      const long index = find_index( end - 1 );
+      if( index >= 0 && domain().includes( sblock( index ) ) &&
+          sblock( index ).status() == Sblock::non_trimmed )
+        change_chunk_status( sblock( index ), Sblock::non_scraped );
+      }
     }
   return 0;
   }
@@ -446,7 +510,7 @@ int Rescuebook::scrape_errors()
 
   for( long i = 0; i < sblocks(); )
     {
-    const Sblock sb = sblock( reverse ? sblocks() - i - 1 : i );
+    const Sblock sb = sblock( reverse ? sblocks() - 1 - i : i );
     if( !domain().includes( sb ) )
       { if( ( !reverse && domain() < sb ) || ( reverse && domain() > sb ) )
           break;
@@ -461,10 +525,10 @@ int Rescuebook::scrape_errors()
       pos = b.end();
       int copied_size = 0, error_size = 0;
       const int retval = copy_and_update( b, copied_size, error_size, msg,
-                                          scraping, true );
+                                          scraping, 1, true );
       if( retval ) return retval;
       update_rates();
-      if( error_size > 0 && exit_on_error ) { e_code |= 2; return 1; }
+      if( error_size > 0 && pause_on_error > 0 ) do_pause_on_error();
       if( !update_mapfile( odes_ ) ) return -2;
       }
     }
@@ -479,18 +543,24 @@ int Rescuebook::copy_errors()
   {
   char msgbuf[80] = "Retrying bad sectors... Retry ";
   const int msglen = std::strlen( msgbuf );
+  bool resume = ( current_status() == retrying );
+  const int first_pass =
+    ( !resume || unidirectional || current_pass() & 1 ) ? 1 : 2;  // odd:even
   bool forward = !reverse;
+  if( !unidirectional && first_pass == 2 ) forward = !forward;
 
-  for( int retry = 1; max_retries < 0 || retry <= max_retries; ++retry )
+  for( int pass = first_pass;
+       max_retries < 0 || pass - first_pass + 1 <= max_retries; ++pass )
     {
     first_post = true;
     snprintf( msgbuf + msglen, ( sizeof msgbuf ) - msglen, "%d %s",
-              retry, forward ? "(forwards)" : "(backwards)" );
-    int retval = forward ? fcopy_errors( msgbuf, retry ) :
-                           rcopy_errors( msgbuf, retry );
+              pass - first_pass + 1, forward ? "(forwards)" : "(backwards)" );
+    int retval = forward ? fcopy_errors( msgbuf, pass, resume ) :
+                           rcopy_errors( msgbuf, pass, resume );
     if( retval != -3 ) return retval;
+    resume = false;
     if( !unidirectional ) forward = !forward;
-    if( retry >= INT_MAX ) break;
+    if( pass >= INT_MAX / 2 ) break;
     }
   return 0;
   }
@@ -499,13 +569,13 @@ int Rescuebook::copy_errors()
 // Return values: 1 I/O error, 0 OK, -1 interrupted, -2 mapfile error.
 // Try to read forwards the damaged areas, one sector at a time.
 //
-int Rescuebook::fcopy_errors( const char * const msg, const int retry )
+int Rescuebook::fcopy_errors( const char * const msg, const int pass,
+                              const bool resume )
   {
   long long pos = 0;
   bool block_found = false;
 
-  if( retry == 1 && current_status() == retrying &&
-      domain().includes( current_pos() ) )
+  if( resume && domain().includes( current_pos() ) )
     {
     Block b( current_pos(), 1 );
     find_chunk( b, Sblock::bad_sector, domain(), hardbs() );
@@ -521,10 +591,10 @@ int Rescuebook::fcopy_errors( const char * const msg, const int retry )
     block_found = true;
     int copied_size = 0, error_size = 0;
     const int retval = copy_and_update( b, copied_size, error_size, msg,
-                                        retrying, true );
+                                        retrying, pass, true );
     if( retval ) return retval;
     update_rates();
-    if( error_size > 0 && exit_on_error ) { e_code |= 2; return 1; }
+    if( error_size > 0 && pause_on_error > 0 ) do_pause_on_error();
     if( !update_mapfile( odes_ ) ) return -2;
     }
   if( !block_found ) return 0;
@@ -535,13 +605,13 @@ int Rescuebook::fcopy_errors( const char * const msg, const int retry )
 // Return values: 1 I/O error, 0 OK, -1 interrupted, -2 mapfile error.
 // Try to read backwards the damaged areas, one sector at a time.
 //
-int Rescuebook::rcopy_errors( const char * const msg, const int retry )
+int Rescuebook::rcopy_errors( const char * const msg, const int pass,
+                              const bool resume )
   {
   long long end = LLONG_MAX;
   bool block_found = false;
 
-  if( retry == 1 && current_status() == retrying &&
-      domain().includes( current_pos() - 1 ) )
+  if( resume && domain().includes( current_pos() - 1 ) )
     {
     Block b( current_pos() - 1, 1 );
     rfind_chunk( b, Sblock::bad_sector, domain(), hardbs() );
@@ -557,10 +627,10 @@ int Rescuebook::rcopy_errors( const char * const msg, const int retry )
     block_found = true;
     int copied_size = 0, error_size = 0;
     const int retval = copy_and_update( b, copied_size, error_size, msg,
-                                        retrying, false );
+                                        retrying, pass, false );
     if( retval ) return retval;
     update_rates();
-    if( error_size > 0 && exit_on_error ) { e_code |= 2; return 1; }
+    if( error_size > 0 && pause_on_error > 0 ) do_pause_on_error();
     if( !update_mapfile( odes_ ) ) return -2;
     }
   if( !block_found ) return 0;
@@ -568,7 +638,9 @@ int Rescuebook::rcopy_errors( const char * const msg, const int retry )
   }
 
 
-void Rescuebook::update_rates( const bool force )
+// Returns true if slow read.
+//
+bool Rescuebook::update_rates( const bool force )
   {
   if( t0 == 0 )
     {
@@ -577,7 +649,7 @@ void Rescuebook::update_rates( const bool force )
     rates_updated = true;
     if( verbosity >= 0 )
       {
-      std::fputs( "\n\n\n\n\n", stdout );
+      std::fputs( "\n\n\n\n\n\n", stdout );
       if( preview_lines > 0 )
         for( int i = -2; i < preview_lines; ++i ) std::fputc( '\n', stdout );
       }
@@ -588,30 +660,50 @@ void Rescuebook::update_rates( const bool force )
     { sleep( 1 ); t2 = std::time( 0 ); }
   if( t2 < t1 )					// clock jumped back
     {
-    const long delta = std::min( t0, t1 - t2 );
+    const long delta = std::min( t0 - 1, t1 - t2 );
     t0 -= delta;
     ts -= delta;
     t1 = t2;
     }
-  if( force && t2 <= t1 ) t2 = t1 + 1;		// force update of e_code
+  const bool force_update = ( force && t2 <= t1 );
+  if( force_update ) t2 = t1 + 1;		// force update of e_code
   if( t2 > t1 )
     {
+    if( tp > 0 )
+      {
+      const long delta = std::min( t0 - 1, (long)tp.round() );
+      t0 -= delta;
+      t1 -= delta;
+      ts -= delta;
+      tp = 0;
+      }
     a_rate = ( finished_size - first_size ) / ( t2 - t0 );
     c_rate = ( finished_size - last_size ) / ( t2 - t1 );
     if( !( e_code & 4 ) )
       {
       if( finished_size != last_size ) { last_size = finished_size; ts = t2; }
-      else if( timeout >= 0 && t2 - ts > timeout && t1 > t0 ) e_code |= 4;
+      else if( !force_update && timeout >= 0 && t2 - ts > timeout && t1 > t0 )
+        e_code |= 4;
       }
-    if( max_error_rate >= 0 && !( e_code & 1 ) )
+    if( !( e_code & 1 ) )
       {
-      error_rate /= ( t2 - t1 );
-      if( error_rate > max_error_rate ) e_code |= 1;
-      else error_rate = 0;
+      error_rate = error_sum / ( t2 - t1 );
+      error_sum = 0;
+      if( max_error_rate >= 0 && error_rate > max_error_rate ) e_code |= 1;
       }
-    t1 = t2;
     rates_updated = true;
+    if( !force_update )
+      {
+      t1 = t2;
+      prev_slow = current_slow;
+      current_slow = ( t1 - t0 > delay_slow &&	// delay checking slow reads
+                       ( ( min_read_rate > 0 && c_rate < min_read_rate ) ||
+                         ( min_read_rate == 0 && c_rate < a_rate / 10 ) ) );
+      if( !current_slow && reset_slow ) slow_reads = 0;
+      return current_slow;
+      }
     }
+  return false;
   }
 
 
@@ -625,7 +717,7 @@ void Rescuebook::show_status( const long long ipos, const char * const msg,
     {
     if( verbosity >= 0 )
       {
-      std::printf( "\r%s%s%s%s%s", up, up, up, up, up );
+      std::printf( "\r%s%s%s%s%s%s", up, up, up, up, up, up );
       if( preview_lines > 0 )
         {
         for( int i = -2; i < preview_lines; ++i ) std::fputs( up, stdout );
@@ -657,22 +749,26 @@ void Rescuebook::show_status( const long long ipos, const char * const msg,
       std::printf( "     opos: %9sB, non-scraped: %9sB,  average rate: %8sB/s\n",
                    format_num( last_ipos + offset() ),
                    format_num( non_scraped_size ), format_num( a_rate, 99999 ) );
-      std::printf( "non-tried: %9sB,     errsize: %9sB,      run time: %11s\n",
-                   format_num( non_tried_size ),
-                   format_num( bad_sector_size, 99999 ), format_time( t1 - t0 ) );
+      std::printf( "non-tried: %9sB,  bad-sector: %9sB,    error rate: %8sB/s\n",
+                   format_num( non_tried_size ), format_num( bad_size ),
+                   format_num( error_rate, 99999 ) );
+      std::printf( "  rescued: %9sB,   bad areas: %8lu,        run time: %11s\n",
+                   format_num( finished_size ), bad_areas,
+                   format_time( t1 - t0 ) );
       if( first_post ) sliding_avg.reset();
       else sliding_avg.add_term( c_rate );
       const long long s_rate = domain().full() ? 0 : sliding_avg();
       const long remaining_time = ( s_rate <= 0 ) ? -1 :
         std::min( std::min( (long long)LONG_MAX, 315359999968464000LL ),
                   ( non_tried_size + non_trimmed_size + non_scraped_size +
-                    ( max_retries ? bad_sector_size : 0 ) + s_rate - 1 ) / s_rate );
-      std::printf( "  rescued: %9sB,      errors: %8ld,  remaining time: %11s\n",
-                   format_num( finished_size ), errors,
+                    ( max_retries ? bad_size : 0 ) + s_rate - 1 ) / s_rate );
+      std::printf( "pct rescued:  %s, read errors:%9lu,  remaining time: %11s\n",
+                   percent_rescued(), read_errors,
                    format_time( remaining_time, remaining_time >= 180 ) );
-      std::printf( "percent rescued: %s      time since last successful read: %11s\n",
-                   format_percentage( finished_size, domain().in_size(), 3, 2 ),
-                   format_time( t1 - ts ) );
+      if( min_read_rate >= -1 ) std::printf( " slow reads:%9lu,", slow_reads );
+      else std::fputs( "                      ", stdout );
+      std::printf( "        time since last successful read: %11s\n",
+                   format_time( ( ts > t0 ) ? t1 - ts : -1 ) );
       if( msg && msg[0] && !errors_or_timeout() )
         {
         const int len = std::strlen( msg ); std::printf( "\r%s", msg );
@@ -681,8 +777,8 @@ void Rescuebook::show_status( const long long ipos, const char * const msg,
         }
       std::fflush( stdout );
       }
-    rate_logger.print_line( t1 - t0, last_ipos, a_rate, c_rate, errors,
-                            bad_sector_size );
+    rate_logger.print_line( t1 - t0, last_ipos, a_rate, c_rate, bad_areas,
+                            bad_size );
     if( !force && !first_post ) read_logger.print_time( t1 - t0 );
     rates_updated = false;
     first_post = false;
@@ -690,35 +786,44 @@ void Rescuebook::show_status( const long long ipos, const char * const msg,
   }
 
 
-Rescuebook::Rescuebook( const long long offset, const long long isize,
+Rescuebook::Rescuebook( const long long offset, const long long insize,
                         Domain & dom, const Domain * const test_dom,
-                        const Rb_options & rb_opts, const char * const iname,
-                        const char * const mapname, const int cluster,
-                        const int hardbs, const bool synchronous )
-  : Mapbook( offset, isize, dom, mapname, cluster, hardbs, rb_opts.complete_only ),
+                        const Mb_options & mb_opts, const Rb_options & rb_opts,
+                        const char * const iname, const char * const mapname,
+                        const int cluster, const int hardbs,
+                        const bool synchronous )
+  : Mapbook( offset, insize, dom, mb_opts, mapname, cluster, hardbs,
+             rb_opts.complete_only ),
     Rb_options( rb_opts ),
     error_rate( 0 ),
+    error_sum( 0 ),
     sparse_size( sparse ? 0 : -1 ),
     non_tried_size( 0 ),
     non_trimmed_size( 0 ),
     non_scraped_size( 0 ),
-    bad_sector_size( 0 ),
+    bad_size( 0 ),
     finished_size( 0 ),
     test_domain( test_dom ),
     iname_( iname ),
+    read_errors( 0 ),
+    slow_reads( 0 ),
     e_code( 0 ),
     synchronous_( synchronous ),
     voe_ipos( -1 ), voe_buf( new uint8_t[hardbs] ),
     a_rate( 0 ), c_rate( 0 ), first_size( 0 ), last_size( 0 ),
-    iobuf_ipos( -1 ), last_ipos( 0 ), t0( 0 ), t1( 0 ), ts( 0 ), oldlen( 0 ),
-    rates_updated( false ), sliding_avg( 30 ), first_post( false ),
+    iobuf_ipos( -1 ), last_ipos( 0 ), t0( 0 ), t1( 0 ), ts( 0 ), tp( 0 ),
+    oldlen( 0 ), rates_updated( false ), current_slow( false ),
+    prev_slow( false ), sliding_avg( 30 ), first_post( false ),
     first_read( true )
   {
   if( preview_lines > softbs() / 16 ) preview_lines = softbs() / 16;
-  const long long csize = isize / 100;
-  if( isize > 0 && skipbs > 0 && max_skipbs == Rb_options::max_max_skipbs &&
+  if( skipbs < 0 )
+    skipbs = round_up( std::max( insize / 100000, (long long)min_skipbs ),
+                       min_skipbs );
+  const long long csize = insize / 100;
+  if( insize > 0 && skipbs > 0 && max_skipbs == max_max_skipbs &&
       csize < max_skipbs )
-    max_skipbs = std::max( (long long)skipbs, csize );
+    max_skipbs = std::max( skipbs, csize );
   skipbs = round_up( skipbs, hardbs );		// make multiple of hardbs
   max_skipbs = round_up( max_skipbs, hardbs );
 
@@ -742,8 +847,8 @@ Rescuebook::Rescuebook( const long long offset, const long long isize,
           sb.status() == Sblock::non_trimmed )
         change_sblock_status( index, Sblock::non_tried );
       }
-  initialize_sizes();				// counts errors
-  if( new_errors_only ) max_errors += errors;
+  initialize_sizes();				// counts bad_areas
+  if( new_bad_areas_only ) max_bad_areas += bad_areas;
   }
 
 
@@ -772,12 +877,13 @@ int Rescuebook::do_rescue( const int ides, const int odes )
           std::printf( " last block size: %9sB\n",
                        format_num( sblock( sblocks() - 1 ).size() ) );
         }
-      if( domain().pos() > 0 || domain().end() < mapfile_isize() )
-        std::printf( "(sizes below are limited to the domain %sB to %sB)\n",
-                     format_num( domain().pos() ), format_num( domain().end() ) );
-      std::printf( "  rescued: %9sB,     errsize: %9sB,  errors: %7ld\n\n",
+      if( domain().pos() > 0 || domain().end() < mapfile_insize() )
+        std::printf( "(sizes limited to domain %lld B to %lld B of %lld B)\n",
+                     domain().pos(), domain().end(), mapfile_insize() );
+      std::printf( "rescued: %sB, tried: %sB, bad-sector: %sB, bad areas: %lu\n\n",
                    format_num( finished_size ),
-                   format_num( bad_sector_size, 99999 ), errors );
+                   format_num( non_trimmed_size + non_scraped_size + bad_size ),
+                   format_num( bad_size ), bad_areas );
       std::fputs( "Current status\n", stdout );
       }
     }
@@ -797,21 +903,27 @@ int Rescuebook::do_rescue( const int ides, const int odes )
   const bool signaled = ( retval == -1 );
   if( signaled ) retval = 0;
   if( retval == 0 && errors_or_timeout() ) retval = 1;
-  if( verbosity >= 0 )
+  if( verbosity >= 0 || event_logger.active() )
     {
-    if( retval == -2 ) std::fputs( "\nMapfile error", stdout );
+    if( retval == -2 ) event_logger.echo_msg( "Mapfile error" );
     else if( retval == 0 && signaled )
-      std::fputs( "\nInterrupted by user", stdout );
+      event_logger.echo_msg( "Interrupted by user" );
     else
       {
       if( e_code & 1 )
-        std::printf( "\nToo high error rate reading input file (%sB/s)",
-                     format_num( error_rate ) );
-      if( e_code & 2 ) std::fputs( "\nToo many errors in input file", stdout );
-      if( e_code & 4 ) std::fputs( "\nTimeout expired", stdout );
+        {
+        char buf[80];
+        snprintf( buf, sizeof buf, "Too high error rate reading input file (%sB/s)",
+                  format_num( error_rate ) );
+        event_logger.echo_msg( buf );
+        }
+      if( e_code & 2 ) event_logger.echo_msg( "Too many bad areas in input file" );
+      if( e_code & 4 ) event_logger.echo_msg( "Timeout expired" );
+      if( e_code & 16 ) event_logger.echo_msg( "Too many read errors" );
+      if( e_code & 32 ) event_logger.echo_msg( "Too many slow reads" );
       }
-    std::fputc( '\n', stdout );
-    std::fflush( stdout );
+    if( verbosity >= 0 )
+      { std::fputc( '\n', stdout ); std::fflush( stdout ); }
     }
   if( retval == -2 ) retval = 1;		// mapfile error
   else
@@ -825,12 +937,14 @@ int Rescuebook::do_rescue( const int ides, const int odes )
     compact_sblock_vector();
     if( !update_mapfile( odes_, true ) && retval == 0 ) retval = 1;
     }
-  if( final_msg().size() )
-    { if( final_errno() ) show_error( final_msg().c_str(), final_errno() );
-      else { std::fputs( final_msg().c_str(), stdout ); std::fputc( '\n', stdout ); } }
+  if( final_msg().size() ) show_error( final_msg().c_str(), final_errno() );
   if( close( odes_ ) != 0 )
-    { show_error( "Can't close outfile", errno );
+    { show_error( "Error closing outfile", errno );
       if( retval == 0 ) retval = 1; }
+  event_logger.print_eor( t1 - t0, percent_rescued(), current_pos(),
+                          status_name( current_status() ) );
+  if( !event_logger.close_file() )
+    show_error( "warning: Error closing the events logging file." );
   if( !rate_logger.close_file() )
     show_error( "warning: Error closing the rates logging file." );
   if( !read_logger.close_file() )
