@@ -1,5 +1,5 @@
 /*  GNU ddrescue - Data recovery tool
-    Copyright (C) 2004-2017 Antonio Diaz Diaz.
+    Copyright (C) 2004-2019 Antonio Diaz Diaz.
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -73,6 +73,7 @@ struct Rb_options
   bool reset_slow;
   bool retrim;
   bool reverse;
+  bool same_file;
   bool simulated_poe;
   bool sparse;
   bool try_again;
@@ -89,8 +90,8 @@ struct Rb_options
       timeout( -1 ), complete_only( false ), new_bad_areas_only( false ),
       noscrape( false ), notrim( false ), reopen_on_error( false ),
       reset_slow( false ), retrim( false ), reverse( false ),
-      simulated_poe( false ), sparse( false ), try_again( false ),
-      unidirectional( false ), verify_on_error( false )
+      same_file( false ), simulated_poe( false ), sparse( false ),
+      try_again( false ), unidirectional( false ), verify_on_error( false )
       {}
 
   bool operator==( const Rb_options & o ) const
@@ -115,6 +116,7 @@ struct Rb_options
                reopen_on_error == o.reopen_on_error &&
                reset_slow == o.reset_slow &&
                retrim == o.retrim && reverse == o.reverse &&
+               same_file == o.same_file &&
                simulated_poe == o.simulated_poe &&
                sparse == o.sparse && try_again == o.try_again &&
                unidirectional == o.unidirectional &&
@@ -182,13 +184,16 @@ class Rescuebook : public Mapbook, public Rb_options
   bool update_rates( const bool force = false );
   void show_status( const long long ipos, const char * const msg = 0,
                     const bool force = false );
+  int copy_command( const char * const command );
+  int status_command( const char * const command ) const;
 public:
-  Rescuebook( const long long offset, const long long isize,
+  Rescuebook( const long long offset, const long long insize,
               Domain & dom, const Domain * const test_dom,
               const Mb_options & mb_opts, const Rb_options & rb_opts,
               const char * const iname, const char * const mapname,
               const int cluster, const int hardbs, const bool synchronous );
   ~Rescuebook() { delete[] voe_buf; }
 
+  int do_commands( const int ides, const int odes );
   int do_rescue( const int ides, const int odes );
   };
